from unittest import TestCase
from pyspark.sql import SparkSession
from pyspark.sql.functions import lit, array
from pii_anonymizer.spark.anonymize.anonymizer import Anonymizer


class TestReplaceAnonymizer(TestCase):
    def setUp(self) -> None:
        self.SPARK = (
            SparkSession.builder.master("local")
            .appName("Test PIIDetector")
            .getOrCreate()
        )

    def test_replace_for_single_analyzer_result(self):
        replace_string = "[REPLACED]"
        test_data_frame = self.SPARK.createDataFrame(
            [("text containing pii", "something else")], ["test"]
        )
        pii_list = ["pii"]
        resultDf = test_data_frame.withColumn(
            "pii_list", array(*map(lit, pii_list))
        ).withColumn("replace_string", lit(replace_string))
        resultDf = resultDf.withColumn(
            "test",
            Anonymizer.replace("test", "replace_string", "pii_list"),
        )

        actual = resultDf.collect()[0][0]

        self.assertEqual(actual, f"text containing {replace_string}")

    def test_replace_for_multiple_analyzer_results(self):
        replace_string = "[REPLACED]"
        test_data_frame = self.SPARK.createDataFrame(
            [("text containing pii1 and pii2", "something else")], ["test"]
        )
        pii_list = ["pii1", "pii2"]
        resultDf = test_data_frame.withColumn(
            "pii_list", array(*map(lit, pii_list))
        ).withColumn("replace_string", lit(replace_string))
        resultDf = resultDf.withColumn(
            "test",
            Anonymizer.replace("test", "replace_string", "pii_list"),
        )
        actual = resultDf.collect()[0][0]
        self.assertEqual(
            actual, f"text containing {replace_string} and {replace_string}"
        )



from unittest import TestCase
from pii_anonymizer.standalone.anonymize.anonymizer import Anonymizer
from pii_anonymizer.standalone.analyze.utils.analyzer_result import AnalyzerResult
from hashlib import sha256


class TestRedactAnonymizer(TestCase):
    def test_hash_for_single_analyzer_result(self):
        text = "text containing pii"
        hashed = sha256("pii".encode("utf-8")).hexdigest()
        analyzer_results = [AnalyzerResult("pii", "PII_DETECTOR", 16, 18)]
        result = Anonymizer.hash(text, analyzer_results)
        self.assertEqual(result, f"text containing {hashed}")

    def test_hash_for_multiple_analyzer_results(self):
        text = "text containing pii1 and pii2"
        hashed1 = sha256("pii1".encode("utf-8")).hexdigest()
        hashed2 = sha256("pii2".encode("utf-8")).hexdigest()
        analyzer_results = [
            AnalyzerResult("pii1", "PII_DETECTOR", 16, 19),
            AnalyzerResult("pii2", "PII_DETECTOR", 25, 28),
        ]
        result = Anonymizer.hash(text, analyzer_results)
        self.assertEqual(result, f"text containing {hashed1} and {hashed2}")



from pii_anonymizer.standalone.analyze.utils.analyzer_result import AnalyzerResult

from hashlib import sha256


class Anonymizer:
    @staticmethod
    def replace(text: str, replace_string: str, analyzer_results: [AnalyzerResult]):
        for result in analyzer_results:
            text = text.replace(result.text, replace_string)
        return text

    @staticmethod
    def hash(text: str, analyzer_results: [AnalyzerResult]):
        for result in analyzer_results:
            text = text.replace(
                result.text, sha256(result.text.encode("utf-8")).hexdigest()
            )
        return text



from unittest import TestCase
from unittest.mock import MagicMock, call
from pyspark.sql import SparkSession
from pii_anonymizer.spark.write.output_writer import (
    OutputWriter,
    output_file_format_err_msg,
)


class TestOutputWriter(TestCase):
    def setUp(self) -> None:
        self.SPARK = (
            SparkSession.builder.master("local").appName("Test CsvWriter").getOrCreate()
        )

    def test_invalid_config_gets_caught_during_initialization(self):
        context = {}
        with self.assertRaises(ValueError) as ve:
            OutputWriter(self.SPARK, config=context)
        self.assertEqual(
            str(ve.exception),
            "Config 'output_file_path' needs to be provided for parsing",
        )

    def test_invalid_output_format_gets_caught_during_initialization(self):
        context = context = {
            "acquire": {"file_path": "/anonymizer/test_data.csv", "delimiter": ","},
            "anonymize": {
                "output_file_path": "/anonymizer/output",
                "output_file_format": "invalid_format",
            },
        }
        with self.assertRaises(ValueError) as ve:
            OutputWriter(self.SPARK, config=context)
        self.assertEqual(str(ve.exception), output_file_format_err_msg)

    def test_correct_output_path_is_generated(self):
        context = {
            "acquire": {"file_path": "/anonymizer/test_data.csv", "delimiter": ","},
            "anonymize": {"output_file_path": "/anonymizer/output"},
        }
        input_file_name = "test_data"
        output_directory = "/anonymizer/output"
        expected = f"{output_directory}/{input_file_name}_anonymized"
        writer = OutputWriter(spark=self.SPARK, config=context)
        self.assertEqual(writer.get_output_file_path(), expected)

    def test_writer_call_correct_methods_on_write(self):
        output_format_list = ["csv", "parquet"]
        for output_format in output_format_list:
            with self.subTest():
                context = {
                    "acquire": {
                        "file_path": "/anonymizer/test_data.csv",
                        "delimiter": ",",
                    },
                    "anonymize": {
                        "output_file_path": "./output",
                        "output_file_format": output_format,
                    },
                }
                df = MagicMock()
                OutputWriter(self.SPARK, config=context).write(df)
                match output_format:
                    case "csv":
                        kall = (
                            call.write.mode("overwrite")
                            .option("header", "true")
                            .csv("./output/test_data_anonymized")
                        )
                        self.assertEqual(df.mock_calls, kall.call_list())
                    case "parquet":
                        kall = (
                            call.write.mode("overwrite")
                            .option("header", "true")
                            .parquet("./output/test_data_anonymized")
                        )
                        self.assertEqual(df.mock_calls, kall.call_list())



ACQUIRE = "acquire"
ANALYZE = "analyze"
ANONYMIZE = "anonymize"
FILE_PATH = "file_path"
REPORT = "report"
LOCATION = "location"
REPORT_LEVEL = "level"
OUTPUT_FILE_PATH = "output_file_path"
OUTPUT_FILE_FORMAT = "output_file_format"



from unittest import TestCase
from unittest.mock import patch
import pandas as pd
from pii_anonymizer.standalone.write.output_writer import (
    OutputWriter,
    output_file_format_err_msg,
)


class TestOutputWriter(TestCase):

    # TODO: check acquire file path exists
    def test_invalid_config_gets_caught_during_initialization(self):
        context = {}
        with self.assertRaises(ValueError) as ve:
            OutputWriter(config=context)
        self.assertEqual(
            str(ve.exception),
            "Config 'output_file_path' needs to be provided for parsing",
        )

    def test_invalid_output_format_gets_caught_during_initialization(self):
        context = {
            "acquire": {"file_path": "/anonymizer/test_data.csv", "delimiter": ","},
            "anonymize": {
                "output_file_path": "/anonymizer/output",
                "output_file_format": "invalid_format",
            },
        }
        with self.assertRaises(ValueError) as ve:
            OutputWriter(config=context)
        self.assertEqual(
            str(ve.exception),
            output_file_format_err_msg,
        )

    def test_correct_output_path_is_generated(self):
        output_format_list = ["csv", "parquet"]
        for output_format in output_format_list:
            with self.subTest():
                context = {
                    "acquire": {
                        "file_path": "/anonymizer/test_data.csv",
                        "delimiter": ",",
                    },
                    "anonymize": {
                        "output_file_path": "/anonymizer/output",
                        "output_file_format": output_format,
                    },
                }
                input_file_name = "test_data"
                output_directory = "/anonymizer/output"
                expected = (
                    f"{output_directory}/{input_file_name}_anonymized.{output_format}"
                )
                writer = OutputWriter(config=context)
                self.assertEqual(writer.get_output_file_path(), expected)

    @patch("pandas.DataFrame.to_csv")
    @patch("pandas.DataFrame.to_parquet")
    def test_writer_call_correct_methods_on_write(self, mock_to_parquet, mock_to_csv):
        df = pd.read_csv("./test_data/test_data.csv")
        mock_to_csv.return_value = None
        mock_to_parquet.return_value = None

        output_format_list = ["csv", "parquet"]
        for output_format in output_format_list:
            with self.subTest():
                context = {
                    "acquire": {
                        "file_path": "/anonymizer/test_data.csv",
                        "delimiter": ",",
                    },
                    "anonymize": {
                        "output_file_path": "./output",
                        "output_file_format": output_format,
                    },
                }
                OutputWriter(config=context).write(df)
                match output_format:
                    case "csv":
                        mock_to_csv.assert_called()
                    case "parquet":
                        mock_to_parquet.assert_called()



from unittest import TestCase
from pyspark.sql import SparkSession
from pyspark.sql.functions import lit, array
from pii_anonymizer.spark.anonymize.anonymizer import Anonymizer
from hashlib import sha256


class TestHashAnonymizer(TestCase):
    def setUp(self) -> None:
        self.SPARK = (
            SparkSession.builder.master("local")
            .appName("Test PIIDetector")
            .getOrCreate()
        )

    def test_hash_for_single_analyzer_result(self):
        test_data_frame = self.SPARK.createDataFrame(
            [("text containing pii", "something else")], ["test"]
        )
        hashed = sha256("pii".encode("utf-8")).hexdigest()
        pii_list = ["pii"]
        resultDf = test_data_frame.withColumn("pii_list", array(*map(lit, pii_list)))
        resultDf = resultDf.withColumn(
            "test",
            Anonymizer.hash("test", "pii_list"),
        )
        actual = resultDf.collect()[0][0]

        self.assertEqual(actual, f"text containing {hashed}")

    def test_hash_for_multiple_analyzer_results(self):
        test_data_frame = self.SPARK.createDataFrame(
            [("text containing pii1 and pii2", "something else")], ["test"]
        )
        pii_list = ["pii1", "pii2"]
        hashed1 = sha256("pii1".encode("utf-8")).hexdigest()
        hashed2 = sha256("pii2".encode("utf-8")).hexdigest()

        resultDf = test_data_frame.withColumn("pii_list", array(*map(lit, pii_list)))
        resultDf = resultDf.withColumn(
            "test",
            Anonymizer.hash("test", "pii_list"),
        )
        actual = resultDf.collect()[0][0]

        self.assertEqual(actual, f"text containing {hashed1} and {hashed2}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pii_anonymizer',
 'pii_anonymizer.common',
 'pii_anonymizer.common.tests',
 'pii_anonymizer.spark',
 'pii_anonymizer.spark.acquire',
 'pii_anonymizer.spark.acquire.tests',
 'pii_anonymizer.spark.analyze',
 'pii_anonymizer.spark.analyze.detectors',
 'pii_anonymizer.spark.analyze.detectors.tests',
 'pii_anonymizer.spark.analyze.utils',
 'pii_anonymizer.spark.anonymize',
 'pii_anonymizer.spark.anonymize.tests',
 'pii_anonymizer.spark.report',
 'pii_anonymizer.spark.report.tests',
 'pii_anonymizer.spark.write',
 'pii_anonymizer.spark.write.tests',
 'pii_anonymizer.standalone',
 'pii_anonymizer.standalone.acquire',
 'pii_anonymizer.standalone.acquire.tests',
 'pii_anonymizer.standalone.analyze',
 'pii_anonymizer.standalone.analyze.detectors',
 'pii_anonymizer.standalone.analyze.detectors.tests',
 'pii_anonymizer.standalone.analyze.utils',
 'pii_anonymizer.standalone.analyze.utils.tests',
 'pii_anonymizer.standalone.anonymize',
 'pii_anonymizer.standalone.anonymize.tests',
 'pii_anonymizer.standalone.report',
 'pii_anonymizer.standalone.report.tests',
 'pii_anonymizer.standalone.tests',
 'pii_anonymizer.standalone.tests.config',
 'pii_anonymizer.standalone.write',
 'pii_anonymizer.standalone.write.tests']

package_data = \
{'': ['*']}

install_requires = \
['fastparquet>=2022.11.0,<2023.0.0',
 'pandas>=1.5.0,<2.0.0',
 'pyspark>=3.3.0,<3.4.0']

setup_kwargs = {
    'name': 'pii-anonymizer',
    'version': '0.2.1',
    'description': 'Data Protection Framework is a python library/command line application for identification, anonymization and de-anonymization of Personally Identifiable Information data.',
    'long_description': '# Data Protection Framework\nData Protection Framework is a python library/command line application for identification, anonymization and de-anonymization of Personally Identifiable Information data.\n\nThe framework aims to work on a two-fold principle for detecting PII:\n1. Using RegularExpressions using a pattern\n2. Using NLP for detecting NER (Named Entity Recognitions)\n\n## Common Usage\n1. `pip install pii-anonymizer`\n2. Specify configs in `pii-anonymizer.json`\n3. Choose whether to run in standalone or spark mode with `python -m pii_anonymizer.standalone` or `python -m pii_anonymizer.spark`\n\n## Features and Current Status\n\n### Completed\n * Following Global detectors have been completed:\n   * [x] EMAIL_ADDRESS :  An email address identifies the mailbox that emails are sent to or from. The maximum length of the domain name is 255 characters, and the maximum length of the local-part is 64 characters.\n   * [x] CREDIT_CARD_NUMBER : A credit card number is 12 to 19 digits long. They are used for payment transactions globally.\n\n * Following detectors specific to Singapore have been completed:\n   * [x] PHONE_NUMBER : A telephone number.\n   * [x] FIN/NRIC : A unique set of nine alpha-numeric characters on the Singapore National Registration Identity Card.\n\n * Following anonymizers have been added\n    * [x] Replacement (\'replace\'): Replaces a detected sensitive value with a specified surrogate value. Leave the value empty to simply delete detected sensitive value.\n    * [x] Hash (\'hash\'): Hash detected sensitive value with sha256.\n\n\n### TO-DO\nFollowing features  are part of the backlog with more features coming soon\n * Detectors:\n    * [ ] NAME\n    * [ ] ADDRESS\n * Anonymizers:\n    * [ ] Encryption :  Encrypts the original sensitive data value using a cryptographic key. Cloud DLP supports several types of tokenization, including transformations that can be reversed, or "re-identified."\n    * [ ] Masking: Replaces a number of characters of a sensitive value with a specified surrogate character, such as a hash (#) or asterisk (*).\n    * [ ] Bucketing: "Generalizes" a sensitive value by replacing it with a range of values. (For example, replacing a specific age with an age range,\n    or temperatures with ranges corresponding to "Hot," "Medium," and "Cold.")\n\n\nYou can have a detailed at upcoming features and backlog in this [Github Board](https://github.com/thoughtworks-datakind/anonymizer/projects/1?fullscreen=true)\n\n## Development setup\n1. Install [Poetry](https://python-poetry.org/docs/#installing-with-the-official-installer)\n2. Setup hooks and install packages with `make install`\n\n### Config JSON\nAn example for the config JSON is located at `<PROJECT_ROOT>/pii-anonymizer.json`\n```\n{\n  "acquire": {\n    "file_path": <FILE PATH TO YOUR INPUT CSV>,\n    "delimiter": <YOUR CSV DELIMITER>\n  },\n  "analyze": {\n    "exclude": [\'Exception\']\n  },\n  "anonymize": {\n    "mode": <replace|hash>,\n    "value": "string to replace",\n    "output_file_path" : <PATH TO YOUR CSV OUTPUT FOLDER>,\n    "output_file_format": <csv|parquet>\n  },\n  "report" : {\n    "location" : <PATH TO YOUR REPORT OUTPUT FOLDER>,\n    "level" : <LOG LEVEL>\n  }\n}\n```\n\n### Running Tests\nYou can run the tests by running `make test` or triggering shell script located at `<PROJECT_ROOT>/bin/run_tests.sh`\n\n### Trying out on local\n\n##### Anonymizing a delimited csv file\n1. Set up a JSON config file similar to the one seen at the project root.\nIn the \'acquire\' section of the json, populate the input file path and the delimiter.\nIn the \'report\' section, provide the output path, where you want the PII detection report to be generated.\nA \'high\' level report just calls out which columns have PII attributes.\nA \'medium\' level report calls out the percentage of PII in each column and the associated PII (email, credit card, etc)type for the same.\n2. Run the main class - `python -m pii_anonymizer.standalone --config <optionally, path of the config file or leave blank to defaults to pii-anonymizer.json>`\nYou should see the report being appended to the file named \'report_\\<date\\>.log\' in the output path specified in the\nconfig file.\n\n### Packaging\nRun `poetry build` and the `.whl` file will be created in the `dist` folder.\n\n### Licensing\nDistributed under the MIT license. See ``LICENSE`` for more information.\n\n### Contributing\n\nYou want to help out? _Awesome_!\n',
    'author': 'Thoughtworks',
    'author_email': 'thoughtworks@thoughtworks.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from pii_anonymizer.common.constants import ANONYMIZE, ANALYZE

analyze_err_msg = f"{ANALYZE} key is required in config"
anonymize_err_msg = f"{ANONYMIZE} key is required in config"
anonymize_mode = ["replace", "hash"]
anonymize_mode_err_msg = f"{ANONYMIZE}'s mode must be {' or '.join(anonymize_mode)}"


def validate(config):
    anonymize_mode_config = config[ANONYMIZE]["mode"]

    if config.get(ANALYZE) is None:
        return ValueError(analyze_err_msg)
    if config.get(ANONYMIZE) is None:
        return ValueError(anonymize_err_msg)

    if anonymize_mode_config != None and anonymize_mode_config not in anonymize_mode:
        raise ValueError(anonymize_mode_err_msg)



import pandas as pd
from pandas import DataFrame

from pii_anonymizer.common.constants import (
    ACQUIRE,
    ANONYMIZE,
    OUTPUT_FILE_PATH,
    FILE_PATH,
    OUTPUT_FILE_FORMAT,
)

output_file_format = ["csv", "parquet"]
output_file_format_err_msg = (
    f"Output file format must be {' or '.join(output_file_format)}"
)


class OutputWriter:
    def __init__(self, config):
        self.__validate_config(config)
        self.__validate_output_format(config)
        self.input_file_name = config[ACQUIRE][FILE_PATH]
        self.output_path = config[ANONYMIZE][OUTPUT_FILE_PATH]

    def __validate_config(self, config):
        if (
            ANONYMIZE not in config
            or not config[ANONYMIZE]
            or OUTPUT_FILE_PATH not in config[ANONYMIZE]
            or not config[ANONYMIZE][OUTPUT_FILE_PATH]
        ):
            raise ValueError(
                "Config 'output_file_path' needs to be provided for parsing"
            )

    def __validate_output_format(self, config):
        self.output_format = config[ANONYMIZE].get(OUTPUT_FILE_FORMAT, "csv")
        if self.output_format not in output_file_format:
            raise ValueError(output_file_format_err_msg)

    def get_output_file_path(self):
        file_name = self.input_file_name.split("/")[-1]
        file_name_no_extension = file_name.split(".")[0]

        result = f"{self.output_path}/{file_name_no_extension}_anonymized.{self.output_format}"
        return result

    def write(self, df: DataFrame):
        match self.output_format:
            case "csv":
                df.to_csv(self.get_output_file_path(), index=False)
            case "parquet":
                df.to_parquet(self.get_output_file_path(), index=False)
        print("Anonymized output has been successfully created!")



from hashlib import sha256
from pyspark.sql.functions import udf
from pyspark.sql.types import StringType


class Anonymizer:
    @udf(returnType=StringType())
    def replace(text: str, replace_string: str, pii_list):
        for word in pii_list:
            text = text.replace(word, replace_string)
        return text

    @udf(returnType=StringType())
    def hash(text: str, pii_list):
        for word in pii_list:
            text = text.replace(word, sha256(word.encode("utf-8")).hexdigest())
        return text
